from enum import StrEnum


class Planet(StrEnum):
    Serpulo = 'serpulo'
    Erekir = 'erekir'


SectorNames = {
    Planet.Serpulo: {
        1: 'Atolls',
        3: 'Testing Grounds',
        15: 'Ground Zero',
        18: 'The Craters',
        20: 'Stained Mountains',
        21: 'Fungal Pass',
        23: 'Tar Fields',
        39: 'Weathered Channels',
        47: 'Sea Port',
        50: 'Frontier',
        54: 'Cruxscape',
        64: 'Facility 32M',
        # 68: 'Polar Aerodrome',
        81: 'Biomass Synthesis Facility',
        86: 'Frozen Forest',
        93: 'Planetary Launch Terminal',
        101: 'Salt Flats',
        108: 'Coastline',
        123: 'Desolate Rift',
        130: 'Nuclear Production Complex',
        134: 'Overgrowth',
        165: 'Extraction Outpost',
        210: 'Infested Canyons',
        213: 'Ruinous Shores',
        216: 'Naval Fortress',
        221: 'Tainted Woods',
        227: 'Impact 0078',
        246: 'Windswept Islands',
        260: 'Mycelia Bastion',
        264: 'Geothermal Stronghold',
    },
    Planet.Erekir: {
        3: 'Crevice',
        5: 'Karst',
        10: 'The Onset',
        12: 'Origin',
        14: 'Atlas',
        18: 'Stronghold',
        19: 'Split',
        25: 'Marsh',
        29: 'Basin',
        30: 'Peaks',
        36: 'Intersect',
        37: 'Crossroads',
        39: 'Ravine',
        41: 'Lake',
        43: 'Caldera',
        58: 'Siege',
        88: 'Aegis',
    },
}

ItemIds = {
    Planet.Serpulo: [
        'blast-compound',
        'coal',
        'copper',
        'graphite',
        'lead',
        'metaglass',
        'phase-fabric',
        'plastanium',
        'pyratite',
        'sand',
        'scrap',
        'silicon',
        'thorium',
        'titanium',
        'spore-pod',
        'surge-alloy',
    ],
    Planet.Erekir: [
        'beryllium',
        'carbide',
        # 'dormant-cyst',
        # 'fissile-matter',
        'graphite',
        'oxide',
        'phase-fabric',
        'thorium',
        'tungsten',
        'silicon',
        'surge-alloy',
    ]
}

ItemColors = {
    'beryllium': '#3a8f64',
    'blast-compound': '#ff795e',
    'carbide': '#89769a',
    'coal': '#53565c', # Real: #272727
    'copper': '#d99d73',
    # 'dormant-cyst': '#df824d',
    # 'fissile-matter': '#5e988d',
    'graphite': '#b2c6d2',
    'lead': '#8c7fa9',
    'metaglass': '#ebeef5',
    'oxide': '#e4ffd6',
    'phase-fabric': '#f4ba6e',
    'plastanium': '#cbd97f',
    'pyratite': '#ffaa5f',
    'scrap': '#777777',
    'sand': '#f7cba4',
    'thorium': '#f9a3c7',
    'titanium': '#8da1e3',
    'tungsten': '#768a9a',
    'silicon': '#53565c',
    'spore-pod': '#7457ce',
    'surge-alloy': '#f3e979',
}


from mindustry_campaign_stats.constants import Planet, SectorNames, ItemIds
from typing import Dict, List, Union, Optional
from datetime import datetime, timezone
import dataclasses
import re


@dataclasses.dataclass
class StorageStats:
    capacity: int
    items: Dict[str, int]


@dataclasses.dataclass
class StorageAndProductionStatsMixin:
    storage: StorageStats
    rawProduction: Dict[str, float]  # Per minute
    netProduction: Dict[str, float]  # Per minute


@dataclasses.dataclass
class SectorStats(StorageAndProductionStatsMixin):
    name: str
    availability: List[str]
    imports: Dict[str, float]  # Per minute
    exports: Dict[str, float]  # Per minute


@dataclasses.dataclass
class TotalsStats(StorageAndProductionStatsMixin):
    pass


@dataclasses.dataclass
class Stats:
    date: datetime
    planet: Planet
    totals: TotalsStats
    sectors: Dict[int, SectorStats]

    def to_dict(self) -> Dict:
        ret = dataclasses.asdict(self)

        ret['date'] = ret['date'].isoformat()
        ret['planet'] = ret['planet'].value

        return ret


class StatsBuilder:
    settings: Dict[str, Union[bool, float, int, bytes, str]]
    planet: Planet
    filter_items: Optional[List[str]] = None
    filter_sectors: Optional[List[str]] = None

    sectors_data: Dict[int, Dict]

    def __init__(self, settings: Dict[str, Union[bool, float, int, bytes, str]], planet: Planet):
        self.settings = settings
        self.planet = planet

        self.sectors_data = self.get_sectors_data()

    def build_sectors(self) -> Dict:
        return {
            sector_id: SectorStats(
                name=SectorNames.get(self.planet).get(sector_id, str(sector_id)),
                availability=sector_info.get('resources', []),
                storage=StorageStats(
                    capacity=sector_info.get('storageCapacity', 0),
                    items=sector_info.get('items', {})
                ),
                rawProduction={
                    item_id: item_info.get('mean', 0) * 60 for item_id, item_info in
                    sector_info.get('rawProduction', {}).items()
                },
                netProduction={
                    item_id: item_info.get('mean', 0) * 60 for item_id, item_info in
                    sector_info.get('production', {}).items()
                },
                imports={
                    item_id: item_info.get('mean', 0) * 60 for item_id, item_info in
                    sector_info.get('imports', {}).items()
                },
                exports={
                    item_id: item_info.get('mean', 0) * 60 for item_id, item_info in
                    sector_info.get('export', {}).items()
                }
            ) for sector_id, sector_info in self.sectors_data.items()
        }

    def build_totals(self) -> TotalsStats:
        return TotalsStats(
            storage=StorageStats(
                capacity=sum([
                    sector_info.get('storageCapacity', 0) for sector_info in self.sectors_data.values()
                ]),
                items={
                    item_id: sum([
                        sector_info.get('items', {}).get(item_id, 0) for sector_info in self.sectors_data.values()
                    ]) for item_id in ItemIds.get(self.planet)
                }
            ),
            rawProduction={
                item_id: sum([
                    sector_info.get('rawProduction', {}).get(item_id, {}).get('mean', 0) * 60 for sector_info in
                    self.sectors_data.values() if sector_info.get('rawProduction', {}).get(item_id, {}).get('mean', 0)
                ]) for item_id in ItemIds.get(self.planet)
            },
            netProduction={
                item_id: sum([
                    sector_info.get('production', {}).get(item_id, {}).get('mean', 0) * 60 for sector_info in
                    self.sectors_data.values() if sector_info.get('production', {}).get(item_id, {}).get('mean', 0)
                ]) for item_id in ItemIds.get(self.planet)
            }
        )

    def get_sectors_data(self) -> Dict[int, Dict]:
        sector_name_regex = re.compile(fr'{self.planet.value}-s-(?P<number>\d+)-info')

        sectors_data = {}

        for key, value in self.settings.items():
            sector_name_match = sector_name_regex.match(key)

            if not sector_name_match:
                continue

            sector_number = int(sector_name_match.groupdict()['number'])

            sectors_data[sector_number] = value

        return sectors_data


def compute(settings: Dict[str, Union[bool, float, int, bytes, str]], planet: Planet) -> Stats:
    builder = StatsBuilder(settings, planet)

    return Stats(
        date=datetime.now(timezone.utc),
        planet=planet,
        totals=builder.build_totals(),
        sectors=builder.build_sectors()
    )



from typing import BinaryIO, Union, Dict
from mutf8 import decode_modified_utf8
from enum import IntEnum
import ubjson
import struct


class SettingType(IntEnum):
    Boolean = 0
    Int = 1
    Long = 2
    Float = 3
    String = 4
    Binary = 5


class SettingsReader:
    fp: BinaryIO

    def __init__(self, fp: BinaryIO):
        self.fp = fp

    def read_mutf8(self) -> str:
        return decode_modified_utf8(
            self.read_bytes(
                self.read_uint16()
            )
        )

    def read_boolean(self) -> bool:
        return self.unpack('?')

    def read_float(self) -> float:
        return self.unpack('f', 4)

    def read_int8(self) -> int:
        return self.unpack('b')

    def read_uint16(self) -> int:
        return self.unpack('H', 2)

    def read_int32(self) -> int:
        return self.unpack('i', 4)

    def read_int64(self) -> int:
        return self.unpack('q', 8)

    def read_bytes(self, size: int) -> bytes:
        return self.fp.read(size)

    def unpack(self, fmt: str, size: int = 1) -> Union[bool, float, int, bytes]:
        ret = struct.unpack(
            f'>{fmt}',
            self.read_bytes(size)
        )

        return ret[0] if len(ret) == 1 else ret


def load(fp: BinaryIO) -> Dict[str, Union[bool, float, int, bytes, str]]:
    settings = {}
    reader = SettingsReader(fp)

    fields_count = reader.read_int32()

    if fields_count <= 0:
        raise ValueError('Invalid settings files: fields count is lower than or equal to 0')

    for _ in range(fields_count):
        field_name = reader.read_mutf8()
        field_type_id = reader.read_int8()

        try:
            field_type = SettingType(field_type_id)
        except ValueError:
            raise ValueError(f'Unhandled field type ID "{field_type_id}" for field "{field_name}"') from None

        if field_type == SettingType.Boolean:
            settings[field_name] = reader.read_boolean()
        elif field_type == SettingType.Int:
            settings[field_name] = reader.read_int32()
        elif field_type == SettingType.Long:
            settings[field_name] = reader.read_int64()
        elif field_type == SettingType.Float:
            settings[field_name] = reader.read_float()
        elif field_type == SettingType.String:
            settings[field_name] = reader.read_mutf8()
        elif field_type == SettingType.Binary:
            settings[field_name] = reader.read_bytes(reader.read_int32())

            # If it looks like ubjson data, try to read it
            if settings[field_name].startswith((b'{', b'[')):
                try:
                    settings[field_name] = ubjson.loadb(settings[field_name])
                except:
                    pass

    if reader.read_bytes(1):
        raise ValueError('Expected EOF, but got something to read')

    return settings
